#!/usr/bin/env python
# Copyright (C) 2018  Nexedi SA and Contributors.
#                     Kirill Smelkov <kirr@nexedi.com>
#
# This program is free software: you can Use, Study, Modify and Redistribute
# it under the terms of the GNU General Public License version 3, or (at your
# option) any later version, as published by the Free Software Foundation.
#
# You can also Link and Combine this program with other software covered by
# the terms of any of the Free Software licenses or any of the Open Source
# Initiative approved licenses and Convey the resulting work. Corresponding
# source of such a combination shall include the source code for all other
# software used.
#
# This program is distributed WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See COPYING file for full licensing terms.
# See https://www.nexedi.com/licensing for rationale and options.
"""Slapns is a playground for SlapOS + Namespaces"""

from unshare import unshare, CLONE_NEWUSER, CLONE_NEWNS, CLONE_NEWPID, CLONE_NEWNET
from os.path import abspath
from subprocess import call
import os, sys, ctypes, errno
import logging as log

MS_RDONLY = 1 << 0
MS_BIND   = 1 << 12

# from https://stackoverflow.com/questions/1667257/how-do-i-mount-a-filesystem-using-python
# XXX better use pylibmount from util-linux directly
def mount(source, target, fs, flags=0, options=None):
  ret = ctypes.CDLL('libc.so.6', use_errno=True).mount(source, target, fs, flags, options)
  if ret < 0:
    errno = ctypes.get_errno()
    raise RuntimeError("Error mounting {} ({}) on {} with options '{}': {}".
     format(source, fs, target, options, os.strerror(errno)))


# bind does `mount --bind source target ...`
def bind(source, target, flags=0, options=''):
    return mount(source, target, None, flags | MS_BIND, options)



# readfile reads data from file @path.
def readfile(path):
    with open(path, "rb") as f:
        return f.read()

# writefile writes data to file @path.
def writefile(path, data):
    with open(path, "wb") as f:
        f.write(data)

# mkdir_p makes sure a directory exists at path.
def mkdir_p(path, mode=0777):
    try:
        os.makedirs(path, mode)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

# ---- uid/gid map setup ----

# On Linux security for setting UID translation from a user namespace to its
# parent is very ad-hoc. Basically single `0 -> UID-in-parent` rule is allowed
# to be setup from a regular users, while everything else requires CAP_SETUID
# in parent user namespace:
#
#   https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/kernel/user_namespace.c?id=v4.19-rc6-177-gcec4de302c5f#n1086
#
# then there is also newuidmap/newgidmap SUID helpers (part of shadow package) which
# read /etc/sub{uid,gid} to see which ID ranges are allowed for a user.
#
# The defaults, when adding a user into a system, is to populate that file with
# 65K ID ranges for every user, for example
#
#   slapuser4:1214112:65536
#
# This way when slapns starts without having CAP_SETUID it tries to use
# newuidmap/newgidmap to setup allowed UID ranges in created namespace.

# subid_allowed finds out which sub uid/gid ranges are allowed for current user.
# -> [] of (id, count)
def subid_allowed(subid_file):
    try:
        f = open(subid_file, "r")
    except IOError as e:
        log.warn("%s: cannot find allowed: %s" % (subid_file, e))
        return []

    mylogin = os.getlogin()
    myuid = "%d" % os.getuid()

    subidv = []
    try:
        for _ in f.readlines():
            # kirr:1279648:65536
            u, id_, count = _.split(":")
            if not (u == mylogin or u == myuid):
                continue

            id_   = int(id_)
            count = int(count)
            subidv.append((id_, count))

        return subidv
    finally:
        f.close()

# idmap_trysetup_viashadow tries to setup uid/gid mapping for a process
# identified by pid via SUID helpers from shadow package.
def idmap_trysetup_viashadow(kind, pid):
    myid = getattr(os, "get"+kind)()
    argv = ["new%smap" % kind, "%d" % pid, "0", "%d" % myid, "1"]  # includes 0 <- uid which is always allowed

    # find out other ID ranges that we can use. Start remapping to [1, ...].
    # We need low IDs to be present, because e.g. glibc.openpty wants to do
    # chown(:tty) and "tty" GID=5.
    x = 1
    for u, count in subid_allowed("/etc/sub%s" % kind):
        argv += ["%d" % x, "%d" % u, "%d" % count]
        x = x + count

    st = call(argv)
    if st != 0:
        log.warn("new%smap failed -> will fallback to `unshare -r` style setup ..." % kind)


# ----------------------------------------


# usage: slapns <slappart>
# create new container inside <slappart> directory + run it.
def main():
    slappart = sys.argv[1]

    unshareonly = slappart.startswith("-U")
    share = set([])  # of path that not to unshare
    if unshareonly:
        _ = slappart[2:]    # -U/tmp:/run -> /tmp:/run
        share = set(_.split(":"))
        slappart = ""
    if not unshareonly:
        # create directories inside container
        dirv = ["/proc", "/sys", "/dev", "/dev/pts",
                "/bin", "/sbin", "/lib", "/lib64", "/usr/bin", "/usr/lib",
                "/etc",
                "/tmp", "/run",
                ]
        for _ in dirv:
            mkdir_p(slappart + _)

    # xmount mounts source to slappart/target
    def xmount(source, target, fs):
        if target in share:
            return  # user asked us not to unshare this
        mount(source, slappart + target, fs)

    # find out my uid/gid
    uid = os.getuid()
    gid = os.getgid()

    # fork a child in which we will call unshare and onto which we'll apply new{uid,gid}map
    unshared_rx, unshared_tx = os.pipe()    # parent <- child   "unshared"
    mapsetup_rx, mapsetup_tx = os.pipe()    # parent -> child   "uid/gid map setup complete"
    pid = os.fork()
    if pid != 0:
        # parent
        os.close(unshared_tx)
        os.close(mapsetup_rx)

        # <- child "unshared"
        _ = os.read(unshared_rx, 1)
        assert _ == "U", `_`

        # TODO check if we have CAP_SETUID and then we can write /proc/.../uid_map directly
        # ^^^ = "slapns inside slapns" case.

        # try to setup uid/git mappings via SUID helper
        idmap_trysetup_viashadow("uid", pid)
        idmap_trysetup_viashadow("gid", pid)

        # signal to child {uid,gid}_map setup complete
        _ = os.write(mapsetup_tx, "M")
        assert _ == 1, _

        # wait on child
        _, st = os.waitpid(pid, 0)
        sys.exit(st >> 8)   # st = (exit << 8) | signal


    # child
    os.close(unshared_rx)
    os.close(mapsetup_tx)

    # unshare from parent namespace (unshare -Umpn)
    unshare(CLONE_NEWUSER | CLONE_NEWNS | CLONE_NEWPID | CLONE_NEWNET)

    # parent <- "unshared"
    _ = os.write(unshared_tx, "U")
    assert _ == 1, _

    # <- parent "uid/gid map setup complete"
    _ = os.read(mapsetup_rx, 1)
    assert _ == "M", `_`

    # update uid/gid maps trivially (uid/gid 0 inside = `unshare -r`) if parent failed.
    if readfile("/proc/self/uid_map") == "":
        writefile("/proc/self/uid_map", "0 %d 1" % uid)
    if readfile("/proc/self/gid_map") == "":
        writefile("/proc/self/setgroups", "deny")
        writefile("/proc/self/gid_map", "0 %d 1" % gid)


    # mount new tmpfs'es
    xmount("none", "/tmp", "tmpfs")
    xmount("none", "/run", "tmpfs")

    # separate instance for pseudo terminals
    # XXX we might need to add gid=5 and other mounting options:
    # XXX see https://lab.nexedi.com/kirr/slapns/commit/1c4ae85d#note_68084
    xmount("none", "/dev/pts", "devpts")

    if not unshareonly:
        # read-only bind mount bin, lib, ... from SR
        # FIXME stub: here we bind from base system for now
        bind("/bin",    slappart + "/bin",  MS_RDONLY)
        bind("/sbin",   slappart + "/sbin", MS_RDONLY)
        bind("/lib",    slappart + "/lib",  MS_RDONLY)
        bind("/lib64",  slappart + "/lib64", MS_RDONLY)
        bind("/usr/bin", slappart + "/usr/bin", MS_RDONLY)
        bind("/usr/lib", slappart + "/usr/lib", MS_RDONLY)

        # XXX we need to setup some small /etc/{passwd,group} - else e.g. screen is not working
        bind("/etc", slappart + "/etc", MS_RDONLY)

    # XXX sysfs and proc are somehow special - mount succeeds only after fork
    pid = os.fork()
    if pid != 0:
        # parent
        _, st = os.waitpid(pid, 0)
        sys.exit(st >> 8)   # st = (exit << 8) | signal

    # child
    xmount("none", "/proc", "proc")
    xmount("none", "/sys", "sysfs")

    # TODO setup networking
    os.system("/sbin/ifconfig lo 127.0.0.1")    # XXX at least loopback works

    # chroot to container
    if not unshareonly:
        slappart = abspath(slappart)
        os.chdir(slappart)
        os.chroot(slappart) # XXX -> try pivot_root

    # FIXME stub: -> $SHELL
    os.execv("/bin/bash", ["bash"])


if __name__ == '__main__':
    main()
